"use strict";
var net = require("net");
var ref = require("ref");
var os = require("os");
var path = require("path");
var util = require("util");
var iosCore = require("./ios-core");
var iOSProxyServices = require("./ios-proxy-services");
var MobileHelper = require("./../mobile-helper");
var hostInfo = require("./../../host-info");
var options = require("./../../options");
var CoreTypes = iosCore.CoreTypes;
var IOSDevice = (function () {
    function IOSDevice(devicePointer, $childProcess, $coreFoundation, $errors, $fs, $injector, $logger, $mobileDevice) {
        this.devicePointer = devicePointer;
        this.$childProcess = $childProcess;
        this.$coreFoundation = $coreFoundation;
        this.$errors = $errors;
        this.$fs = $fs;
        this.$injector = $injector;
        this.$logger = $logger;
        this.$mobileDevice = $mobileDevice;
        this.identifier = null;
        this.voidPtr = ref.refType(ref.types.void);
        this.mountImageCallbackPtr = null;
        this.mountImageCallbackPtr = CoreTypes.am_device_mount_image_callback.toPointer(IOSDevice.mountImageCallback);
    }
    IOSDevice.mountImageCallback = function (dictionary, user) {
        var coreFoundation = $injector.resolve("coreFoundation");
        var logger = $injector.resolve("logger");
        var jsDictionary = coreFoundation.cfTypeTo(dictionary);
        logger.info("[Mounting] %s", jsDictionary["Status"]);
    };
    IOSDevice.prototype.getPlatform = function () {
        return MobileHelper.DevicePlatforms[0 /* iOS */];
    };
    IOSDevice.prototype.getIdentifier = function () {
        if (this.identifier == null) {
            this.identifier = this.$coreFoundation.convertCFStringToCString(this.$mobileDevice.deviceCopyDeviceIdentifier(this.devicePointer));
        }
        return this.identifier;
    };
    IOSDevice.prototype.getDisplayName = function () {
        return this.getValue("ProductType");
    };
    IOSDevice.prototype.getModel = function () {
        return this.getValue("ProductType");
    };
    IOSDevice.prototype.getVersion = function () {
        return this.getValue("ProductVersion");
    };
    IOSDevice.prototype.getVendor = function () {
        return "Apple";
    };
    IOSDevice.prototype.getValue = function (value) {
        this.connect();
        this.startSession();
        try {
            var cfValue = this.$coreFoundation.createCFString(value);
            return this.$coreFoundation.convertCFStringToCString(this.$mobileDevice.deviceCopyValue(this.devicePointer, null, cfValue));
        }
        finally {
            this.stopSession();
            this.disconnect();
        }
    };
    IOSDevice.prototype.validateResult = function (result, error) {
        if (result !== 0) {
            this.$errors.fail(util.format("%s. Result code is: %s", error, result));
        }
    };
    IOSDevice.prototype.isPaired = function () {
        return this.$mobileDevice.deviceIsPaired(this.devicePointer) != 0;
    };
    IOSDevice.prototype.pair = function () {
        var result = this.$mobileDevice.devicePair(this.devicePointer);
        this.validateResult(result, "If your phone is locked with a passcode, unlock then reconnect it");
        return result;
    };
    IOSDevice.prototype.validatePairing = function () {
        var result = this.$mobileDevice.deviceValidatePairing(this.devicePointer);
        this.validateResult(result, "Unable to validate pairing");
        return result;
    };
    IOSDevice.prototype.connect = function () {
        var result = this.$mobileDevice.deviceConnect(this.devicePointer);
        this.validateResult(result, "Unable to connect to device");
        if (!this.isPaired()) {
            this.pair();
        }
        return this.validatePairing();
    };
    IOSDevice.prototype.disconnect = function () {
        var result = this.$mobileDevice.deviceDisconnect(this.devicePointer);
        this.validateResult(result, "Unable to disconnect from device");
    };
    IOSDevice.prototype.startSession = function () {
        var result = this.$mobileDevice.deviceStartSession(this.devicePointer);
        this.validateResult(result, "Unable to start session");
    };
    IOSDevice.prototype.stopSession = function () {
        var result = this.$mobileDevice.deviceStopSession(this.devicePointer);
        this.validateResult(result, "Unable to stop session");
    };
    IOSDevice.prototype.getDeviceValue = function (value) {
        var deviceCopyValue = this.$mobileDevice.deviceCopyValue(this.devicePointer, null, this.$coreFoundation.createCFString(value));
        return this.$coreFoundation.convertCFStringToCString(deviceCopyValue);
    };
    IOSDevice.prototype.lookupApplications = function () {
        var _this = this;
        var func = function () {
            var dictionaryPointer = ref.alloc(CoreTypes.cfDictionaryRef);
            var result = _this.$mobileDevice.deviceLookupApplications(_this.devicePointer, 0, dictionaryPointer);
            if (result !== 0) {
                _this.$errors.fail("Invalid result code %s from device lookup applications.", result);
            }
            var cfDictionary = dictionaryPointer.deref();
            var jsDictionary = _this.$coreFoundation.cfTypeTo(cfDictionary);
            return jsDictionary;
        };
        return this.tryToExecuteFunction(func);
    };
    IOSDevice.prototype.findDeveloperDirectory = function () {
        var _this = this;
        return (function () {
            var childProcess = _this.$childProcess.spawnFromEvent("xcode-select", ["-print-path"], "close").wait();
            return childProcess.stdout.trim();
        }).future()();
    };
    IOSDevice.prototype.tryToExecuteFunction = function (func) {
        this.connect();
        try {
            this.startSession();
            try {
                return func.apply(this, []);
            }
            finally {
                this.stopSession();
            }
        }
        finally {
            this.disconnect();
        }
    };
    IOSDevice.prototype.findDeveloperDiskImageDirectoryPath = function () {
        var _this = this;
        return (function () {
            var developerDirectory = _this.findDeveloperDirectory().wait();
            var buildVersion = _this.getDeviceValue("BuildVersion");
            var productVersion = _this.getDeviceValue("ProductVersion");
            var productVersionParts = productVersion.split(".");
            var productMajorVersion = productVersionParts[0];
            var productMinorVersion = productVersionParts[1];
            var developerDiskImagePath = path.join(developerDirectory, "Platforms", "iPhoneOS.platform", "DeviceSupport");
            var supportPaths = _this.$fs.readDirectory(developerDiskImagePath).wait();
            var supportPath = null;
            _.each(supportPaths, function (sp) {
                var parts = sp.split(' ');
                var version = parts[0];
                var versionParts = version.split(".");
                var supportPathData = {
                    version: version,
                    majorVersion: versionParts[0],
                    minorVersion: versionParts[1],
                    build: parts.length > 1 ? parts[1].replace(/[()]/, function () { return ""; }) : null,
                    path: path.join(developerDiskImagePath, sp)
                };
                if (supportPathData.majorVersion === productMajorVersion) {
                    if (!supportPath) {
                        supportPath = supportPathData;
                    }
                    else {
                        if (supportPathData.minorVersion === productMinorVersion) {
                            if (supportPathData.build === buildVersion) {
                                supportPath = supportPathData;
                            }
                            else {
                                if (supportPath.build !== supportPathData.build || supportPath.build === null) {
                                    supportPath = supportPathData;
                                }
                            }
                        }
                    }
                }
            });
            if (!supportPath) {
                _this.$errors.fail("Unable to find device support path.");
            }
            return supportPath.path;
        }).future()();
    };
    IOSDevice.prototype.mountImage = function () {
        var _this = this;
        return (function () {
            var imagePath = options.ddi;
            if (hostInfo.isWindows()) {
                if (!imagePath) {
                    _this.$errors.fail("On windows operating system you must specify the path to developer disk image using --ddi option");
                }
                var imageSignature = _this.$fs.readFile(util.format("%s.signature", imagePath)).wait();
                var imageSize = _this.$fs.getFsStats(imagePath).wait().size;
                var imageMounterService = _this.startService(iOSProxyServices.MobileServices.MOBILE_IMAGE_MOUNTER);
                var plistService = _this.$injector.resolve(iosCore.PlistService, { service: imageMounterService, format: CoreTypes.kCFPropertyListXMLFormat_v1_0 });
                var result = plistService.exchange({
                    Command: "ReceiveBytes",
                    ImageSize: imageSize,
                    ImageType: "Developer",
                    ImageSignature: imageSignature
                }).wait();
                if (result.Status === "ReceiveBytesAck") {
                    var fileData = _this.$fs.readFile(imagePath).wait();
                    plistService.sendAll(fileData);
                }
                else {
                    var afcService = _this.startService(iOSProxyServices.MobileServices.APPLE_FILE_CONNECTION);
                    var afcClient = _this.$injector.resolve(iOSProxyServices.AfcClient, { service: afcService });
                    afcClient.transfer(imagePath, "PublicStaging/staging.dimage").wait();
                }
                try {
                    result = plistService.exchange({
                        Command: "MountImage",
                        ImageType: "Developer",
                        ImageSignature: imageSignature,
                        ImagePath: "/var/mobile/Media/PublicStaging/staging.dimage"
                    }).wait();
                    if (result.Error) {
                        _this.$errors.fail("Unable to mount image. %s", result.Error);
                    }
                    if (result.Status) {
                        _this.$logger.info("Mount image: %s", result.Status);
                    }
                }
                finally {
                    plistService.close();
                }
            }
            else {
                var func = function () {
                    var developerDiskImageDirectoryPath = _this.findDeveloperDiskImageDirectoryPath().wait();
                    var imagePath = path.join(developerDiskImageDirectoryPath, "DeveloperDiskImage.dmg");
                    _this.$logger.info("Mounting %s", imagePath);
                    var signature = _this.$fs.readFile(util.format("%s.signature", imagePath)).wait();
                    var cfImagePath = _this.$coreFoundation.createCFString(imagePath);
                    var cfOptions = _this.$coreFoundation.cfTypeFrom({
                        ImageType: "Developer",
                        ImageSignature: signature
                    });
                    var result = _this.$mobileDevice.deviceMountImage(_this.devicePointer, cfImagePath, cfOptions, _this.mountImageCallbackPtr);
                    if (result !== 0 && result !== IOSDevice.IMAGE_ALREADY_MOUNTED_ERROR_CODE) {
                        _this.$errors.fail("Unable to mount image on device.");
                    }
                };
                _this.tryToExecuteFunction(func);
            }
        }).future()();
    };
    IOSDevice.prototype.getInterfaceType = function () {
        return this.$mobileDevice.deviceGetInterfaceType(this.devicePointer);
    };
    IOSDevice.prototype.startService = function (serviceName) {
        var _this = this;
        var func = function () {
            var socket = ref.alloc("int");
            var result = _this.$mobileDevice.deviceStartService(_this.devicePointer, _this.$coreFoundation.createCFString(serviceName), socket);
            _this.validateResult(result, "Unable to start service");
            return ref.deref(socket);
        };
        return this.tryToExecuteFunction(func);
    };
    IOSDevice.prototype.deploy = function (packageFile, packageName) {
        var _this = this;
        return (function () {
            var installationProxy = _this.$injector.resolve(iOSProxyServices.InstallationProxyClient, { device: _this });
            installationProxy.deployApplication(packageFile).wait();
            installationProxy.closeSocket();
        }).future()();
    };
    IOSDevice.prototype.debug = function (packageFile, packageName) {
        var _this = this;
        return (function () {
            var debugging = _this.$injector.resolve(IOSDeviceDebugging, {
                packageFile: packageFile,
                packageName: packageName,
                $iOSDevice: _this
            });
            debugging.debug();
        }).future()();
    };
    IOSDevice.prototype.sync = function (localToDevicePaths, appIdentifier, liveSyncUrl, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        return (function () {
            var houseArrestClient = _this.$injector.resolve(iOSProxyServices.HouseArrestClient, { device: _this });
            var afcClientForAppDocuments = houseArrestClient.getAfcClientForAppDocuments(appIdentifier.appIdentifier);
            afcClientForAppDocuments.transferCollection(localToDevicePaths).wait();
            houseArrestClient.closeSocket();
            if (!options.skipRefresh) {
                var afcClientForContainer = houseArrestClient.getAfcClientForAppContainer(appIdentifier.appIdentifier);
                afcClientForContainer.deleteFile("/Library/Preferences/ServerInfo.plist");
                houseArrestClient.closeSocket();
                var notificationProxyClient = _this.$injector.resolve(iOSProxyServices.NotificationProxyClient, { device: _this });
                notificationProxyClient.postNotification("com.telerik.app.refreshWebView");
                notificationProxyClient.closeSocket();
            }
            _this.$logger.out("Successfully synced device with identifier '%s'", _this.getIdentifier());
        }).future()();
    };
    IOSDevice.prototype.openDeviceLogStream = function () {
        var iOSSystemLog = this.$injector.resolve(iOSProxyServices.IOSSyslog, { device: this });
        iOSSystemLog.read();
    };
    IOSDevice.prototype.getInstalledApplications = function () {
        var _this = this;
        return (function () {
            return _(_this.lookupApplications()).keys().sortBy(function (identifier) { return identifier.toLowerCase(); }).value();
        }).future()();
    };
    IOSDevice.prototype.runApplication = function (applicationId) {
        var _this = this;
        return (function () {
            if (hostInfo.isWindows()) {
                _this.$errors.fail("$appbuilder device run command is not supported on Windows for iOS devices.");
            }
            var applications = _this.lookupApplications();
            var application = applications[applicationId];
            if (!application) {
                var sortedKeys = _.sortBy(_.keys(applications));
                _this.$errors.fail("Invalid application id: %s. All available application ids are: %s%s ", applicationId, os.EOL, sortedKeys.join(os.EOL));
            }
            _this.mountImage().wait();
            var gdbServer = _this.$injector.resolve(iosCore.GDBServer, { service: _this.startService(iOSProxyServices.MobileServices.DEBUG_SERVER) });
            var executable = util.format("%s/%s", application.Path, application.CFBundleExecutable);
            gdbServer.run([executable]);
            _this.$logger.info("Successfully run application %s on device with ID %s", applicationId, _this.getIdentifier());
        }).future()();
    };
    IOSDevice.prototype.connectToPort = function (port) {
        var interfaceType = this.getInterfaceType();
        if (interfaceType === IOSDevice.INTERFACE_USB) {
            var connectionId = this.$mobileDevice.deviceGetConnectionId(this.devicePointer);
            var socketRef = ref.alloc(CoreTypes.intType);
            this.$mobileDevice.uSBMuxConnectByPort(connectionId, this.htons(port), socketRef);
            var socketValue = socketRef.deref();
            return new net.Socket({ fd: socketValue });
        }
        return null;
    };
    IOSDevice.prototype.htons = function (port) {
        var result = (port & 0xff00) >> 8 | (port & 0x00ff) << 8;
        return result;
    };
    IOSDevice.IMAGE_ALREADY_MOUNTED_ERROR_CODE = 3892314230;
    IOSDevice.INTERFACE_USB = 1;
    return IOSDevice;
})();
exports.IOSDevice = IOSDevice;
$injector.register("iOSDevice", IOSDevice);
var IOSDeviceDebugging = (function () {
    function IOSDeviceDebugging(packageFile, packageName, $iOSDevice, $injector, $logger, $errors, $npm, $childProcess) {
        this.packageFile = packageFile;
        this.packageName = packageName;
        this.$iOSDevice = $iOSDevice;
        this.$injector = $injector;
        this.$logger = $logger;
        this.$errors = $errors;
        this.$npm = $npm;
        this.$childProcess = $childProcess;
        this.notificationProxyClient = this.$injector.resolve(iOSProxyServices.NotificationProxyClient, { device: this.$iOSDevice });
    }
    IOSDeviceDebugging.prototype.debug = function () {
        if (options["get-port"]) {
            this.$errors.fail({ formatStr: "this feature is not supported in the selected platform", suppressCommandHelp: true });
        }
        else if (options["start"]) {
            this.attachDebugger();
        }
        else if (options["stop"]) {
            this.$errors.fail({ formatStr: "this feature is not supported in the selected platform", suppressCommandHelp: true });
        }
        else if (options["debug-brk"]) {
            this.startAppWithDebugger();
        }
        else {
            this.$logger.info("Should specify exactly one option: debug-brk, start");
        }
    };
    IOSDeviceDebugging.prototype.attachDebugger = function () {
        this.attachAtRuntime();
    };
    IOSDeviceDebugging.prototype.startAppWithDebugger = function () {
        this.$iOSDevice.deploy(this.packageFile, this.packageName).wait();
        this.attachAtStartup();
        this.$iOSDevice.runApplication(this.packageName).wait();
    };
    IOSDeviceDebugging.prototype.attachAtStartup = function () {
        var _this = this;
        this.startTheFrontEnd();
        var appLaunchMessage = this.packageName + ":NativeScript.Debug.AppLaunching";
        this.notificationProxyClient.addObserver(appLaunchMessage, function () {
            _this.$logger.info("Got AppLaunching");
            _this.startTheBackend("WaitForDebugger");
        });
    };
    IOSDeviceDebugging.prototype.attachAtRuntime = function () {
        this.startTheFrontEnd();
        this.startTheBackend("AttachRequest");
    };
    IOSDeviceDebugging.prototype.startTheFrontEnd = function () {
        this.openSafariFrontEnd();
        this.printHowToTerminate();
    };
    IOSDeviceDebugging.prototype.startTheBackend = function (message) {
        this.attachWhenReady();
        this.postDebugNotification(message);
    };
    IOSDeviceDebugging.prototype.openSafariFrontEnd = function () {
        var tnsIosPackage = this.$npm.install("tns-ios").wait();
        var safariPath = path.join(tnsIosPackage, "WebInspectorUI/Safari/Main.html");
        this.$childProcess.exec("open -a Safari " + safariPath).wait();
    };
    IOSDeviceDebugging.prototype.attachWhenReady = function () {
        var _this = this;
        var identifier = this.$iOSDevice.getIdentifier();
        this.$logger.info("Device Identifier: " + identifier);
        var appReadyForAttachMessage = this.packageName + ":NativeScript.Debug.ReadyForAttach";
        this.notificationProxyClient.addObserver(appReadyForAttachMessage, function () {
            _this.$logger.info("Got ReadyForAttach");
            var devicePort = 8080;
            var localPort = 8080;
            var localServer = net.createServer(function (localSocket) {
                _this.$logger.info("Front-end client connected to localhost " + localPort);
                var deviceSocket;
                var buff = new Array();
                localSocket.on('data', function (data) {
                    if (deviceSocket) {
                        deviceSocket.write(data);
                    }
                    else {
                        buff.push(data);
                    }
                });
                localSocket.on('end', function () {
                    _this.$logger.info('Localhost client disconnected!');
                    process.exit(0);
                });
                deviceSocket = _this.$iOSDevice.connectToPort(devicePort);
                _this.$logger.info("Connected localhost " + localPort + " to device " + devicePort);
                buff.forEach(function (data) {
                    deviceSocket.write(data);
                });
                deviceSocket.on('data', function (data) {
                    localSocket.write(data);
                });
                deviceSocket.on('end', function () {
                    _this.$logger.info("Device socket closed!");
                    process.exit(0);
                });
            });
            localServer.listen(localPort, function () {
                _this.$logger.info("Opened localhost " + localPort);
            });
        });
    };
    IOSDeviceDebugging.prototype.printHowToTerminate = function () {
        this.$logger.info("\nSetting up debugger proxy...\n\nPress Ctrl + C to terminate!\n");
    };
    IOSDeviceDebugging.prototype.postDebugNotification = function (notification) {
        var attachRequestMessage = this.packageName + ":NativeScript.Debug.AttachRequest";
        this.notificationProxyClient.postNotification(attachRequestMessage);
        this.$logger.info("Send " + notification);
    };
    return IOSDeviceDebugging;
})();
exports.IOSDeviceDebugging = IOSDeviceDebugging;
$injector.register("iOSDeviceDebugging", IOSDeviceDebugging);
